@@ -125,13 +125,21 @@ contextmanager = contextlib.contextmanager
 
 __all__ = ['start', 'stop', 'reset', 'display', 'profile']
 
-skips = {"util.py:check", "extensions.py:closure",
-         "color.py:colorcmd", "dispatch.py:checkargs",
-         "dispatch.py:<lambda>", "dispatch.py:_runcatch",
-         "dispatch.py:_dispatch", "dispatch.py:_runcommand",
-         "pager.py:pagecmd", "dispatch.py:run",
-         "dispatch.py:dispatch", "dispatch.py:runcommand",
-         "hg.py:<module>", "evolve.py:warnobserrors",
+skips = {
+    r"util.py:check",
+    r"extensions.py:closure",
+    r"color.py:colorcmd",
+    r"dispatch.py:checkargs",
+    r"dispatch.py:<lambda>",
+    r"dispatch.py:_runcatch",
+    r"dispatch.py:_dispatch",
+    r"dispatch.py:_runcommand",
+    r"pager.py:pagecmd",
+    r"dispatch.py:run",
+    r"dispatch.py:dispatch",
+    r"dispatch.py:runcommand",
+    r"hg.py:<module>",
+    r"evolve.py:warnobserrors",
 }
 
 ###########################################################################
@@ -195,11 +203,13 @@ state = ProfileState()
 class CodeSite(object):
     cache = {}
 
-    __slots__ = (u'path', u'lineno', u'function', u'source')
+    __slots__ = (r'path', r'lineno', r'function', r'source')
 
     def __init__(self, path, lineno, function):
+        assert isinstance(path, bytes)
         self.path = path
         self.lineno = lineno
+        assert isinstance(function, bytes)
         self.function = function
         self.source = None
 
@@ -228,7 +238,7 @@ class CodeSite(object):
             lineno = self.lineno - 1
             fp = None
             try:
-                fp = open(self.path)
+                fp = open(self.path, 'rb')
                 for i, line in enumerate(fp):
                     if i == lineno:
                         self.source = line.strip()
@@ -249,8 +259,11 @@ class CodeSite(object):
     def filename(self):
         return os.path.basename(self.path)
 
+    def skipname(self):
+        return r'%s:%s' % (self.filename(), self.function)
+
 class Sample(object):
-    __slots__ = (u'stack', u'time')
+    __slots__ = (r'stack', r'time')
 
     def __init__(self, stack, time):
         self.stack = stack
@@ -261,8 +274,10 @@ class Sample(object):
         stack = []
 
         while frame:
-            stack.append(CodeSite.get(frame.f_code.co_filename, frame.f_lineno,
-                                      frame.f_code.co_name))
+            stack.append(CodeSite.get(
+                pycompat.sysbytes(frame.f_code.co_filename),
+                frame.f_lineno,
+                pycompat.sysbytes(frame.f_code.co_name)))
             frame = frame.f_back
 
         return Sample(stack, time)
@@ -352,14 +367,14 @@ def save_data(path):
     with open(path, 'w+') as file:
         file.write("%f %f\n" % state.accumulated_time)
         for sample in state.samples:
-            time = str(sample.time)
+            time = sample.time
             stack = sample.stack
-            sites = ['\1'.join([s.path, str(s.lineno), s.function])
+            sites = ['\1'.join([s.path, b'%d' % s.lineno, s.function])
                      for s in stack]
-            file.write(time + '\0' + '\0'.join(sites) + '\n')
+            file.write("%d\0%s\n" % (time, '\0'.join(sites)))
 
 def load_data(path):
-    lines = open(path, 'r').read().splitlines()
+    lines = open(path, 'rb').read().splitlines()
 
     state.accumulated_time = [float(value) for value in lines[0].split()]
     state.samples = []
@@ -461,7 +476,7 @@ def display(fp=None, format=3, data=None, **kwargs):
         import sys
         fp = sys.stdout
     if len(data.samples) == 0:
-        print('No samples recorded.', file=fp)
+        fp.write(b'No samples recorded.\n')
         return
 
     if format == DisplayFormats.ByLine:
@@ -482,10 +497,9 @@ def display(fp=None, format=3, data=None, **kwargs):
         raise Exception("Invalid display format")
 
     if format not in (DisplayFormats.Json, DisplayFormats.Chrome):
-        print('---', file=fp)
-        print('Sample count: %d' % len(data.samples), file=fp)
-        print('Total time: %f seconds (%f wall)' % data.accumulated_time,
-              file=fp)
+        fp.write(b'---\n')
+        fp.write(b'Sample count: %d\n' % len(data.samples))
+        fp.write(b'Total time: %f seconds (%f wall)\n' % data.accumulated_time)
 
 def display_by_line(data, fp):
     '''Print the profiler data with each sample line represented
@@ -493,34 +507,34 @@ def display_by_line(data, fp):
     stats = SiteStats.buildstats(data.samples)
     stats.sort(reverse=True, key=lambda x: x.selfseconds())
 
-    print('%5.5s %10.10s   %7.7s  %-8.8s' %
-          ('%  ', 'cumulative', 'self', ''), file=fp)
-    print('%5.5s  %9.9s  %8.8s  %-8.8s' %
-          ("time", "seconds", "seconds", "name"), file=fp)
+    fp.write(b'%5.5s %10.10s   %7.7s  %-8.8s\n' % (
+        b'%  ', b'cumulative', b'self', b''))
+    fp.write(b'%5.5s  %9.9s  %8.8s  %-8.8s\n' % (
+        b"time", b"seconds", b"seconds", b"name"))
 
     for stat in stats:
         site = stat.site
-        sitelabel = '%s:%d:%s' % (site.filename(), site.lineno, site.function)
-        print('%6.2f %9.2f %9.2f  %s' % (stat.selfpercent(),
-                                         stat.totalseconds(),
-                                         stat.selfseconds(),
-                                         sitelabel),
-              file=fp)
+        sitelabel = '%s:%d:%s' % (site.filename(),
+                                  site.lineno,
+                                  site.function)
+        fp.write(b'%6.2f %9.2f %9.2f  %s\n' % (
+            stat.selfpercent(), stat.totalseconds(),
+            stat.selfseconds(), sitelabel))
 
 def display_by_method(data, fp):
     '''Print the profiler data with each sample function represented
     as one row in a table.  Important lines within that function are
     output as nested rows.  Sorted by self-time per line.'''
-    print('%5.5s %10.10s   %7.7s  %-8.8s' %
-          ('%  ', 'cumulative', 'self', ''), file=fp)
-    print('%5.5s  %9.9s  %8.8s  %-8.8s' %
-          ("time", "seconds", "seconds", "name"), file=fp)
+    fp.write(b'%5.5s %10.10s   %7.7s  %-8.8s\n' %
+          ('%  ', 'cumulative', 'self', ''))
+    fp.write(b'%5.5s  %9.9s  %8.8s  %-8.8s\n' %
+          ("time", "seconds", "seconds", "name"))
 
     stats = SiteStats.buildstats(data.samples)
 
     grouped = defaultdict(list)
     for stat in stats:
-        grouped[stat.site.filename() + ":" + stat.site.function].append(stat)
+        grouped[stat.site.filename() + b":" + stat.site.function].append(stat)
 
     # compute sums for each function
     functiondata = []
@@ -545,20 +559,24 @@ def display_by_method(data, fp):
     for function in functiondata:
         if function[3] < 0.05:
             continue
-        print('%6.2f %9.2f %9.2f  %s' % (function[3], # total percent
-                                         function[1], # total cum sec
-                                         function[2], # total self sec
-                                         function[0]), # file:function
-              file=fp)
+        fp.write(b'%6.2f %9.2f %9.2f  %s\n' % (
+            function[3], # total percent
+            function[1], # total cum sec
+            function[2], # total self sec
+            function[0])) # file:function
+
         function[4].sort(reverse=True, key=lambda i: i.selfseconds())
         for stat in function[4]:
             # only show line numbers for significant locations (>1% time spent)
             if stat.selfpercent() > 1:
                 source = stat.site.getsource(25)
+                if sys.version_info.major >= 3 and not isinstance(source, bytes):
+                    source = pycompat.bytestr(source)
+
                 stattuple = (stat.selfpercent(), stat.selfseconds(),
                              stat.site.lineno, source)
 
-                print('%33.0f%% %6.2f   line %s: %s' % (stattuple), file=fp)
+                fp.write(b'%33.0f%% %6.2f   line %d: %s\n' % stattuple)
 
 def display_about_method(data, fp, function=None, **kwargs):
     if function is None:
@@ -592,9 +610,12 @@ def display_about_method(data, fp, function=None, **kwargs):
     parents = [(parent, count) for parent, count in parents.iteritems()]
     parents.sort(reverse=True, key=lambda x: x[1])
     for parent, count in parents:
-        print('%6.2f%%   %s:%s   line %s: %s' %
-            (count / relevant_samples * 100, parent.filename(),
-            parent.function, parent.lineno, parent.getsource(50)), file=fp)
+        fp.write(b'%6.2f%%   %s:%s   line %s: %s\n' %
+            (count / relevant_samples * 100,
+             pycompat.fsencode(parent.filename()),
+             pycompat.sysbytes(parent.function),
+             parent.lineno,
+             pycompat.sysbytes(parent.getsource(50))))
 
     stats = SiteStats.buildstats(data.samples)
     stats = [s for s in stats
@@ -611,23 +632,23 @@ def display_about_method(data, fp, function=None, **kwargs):
         total_self_percent += stat.selfpercent()
         total_cum_percent += stat.totalpercent()
 
-    print(
-        '\n    %s:%s    Total: %0.2fs (%0.2f%%)    Self: %0.2fs (%0.2f%%)\n' %
-        (
-        filename or '___',
-        function,
+    fp.write(
+        b'\n    %s:%s    Total: %0.2fs (%0.2f%%)    Self: %0.2fs (%0.2f%%)\n\n'
+        % (
+        pycompat.sysbytes(filename or '___'),
+        pycompat.sysbytes(function),
         total_cum_sec,
         total_cum_percent,
         total_self_sec,
         total_self_percent
-        ), file=fp)
+        ))
 
     children = [(child, count) for child, count in children.iteritems()]
     children.sort(reverse=True, key=lambda x: x[1])
     for child, count in children:
-        print('        %6.2f%%   line %s: %s' %
+        fp.write(b'        %6.2f%%   line %s: %s\n' %
               (count / relevant_samples * 100, child.lineno,
-               child.getsource(50)), file=fp)
+               pycompat.sysbytes(child.getsource(50))))
 
 def display_hotpath(data, fp, limit=0.05, **kwargs):
     class HotNode(object):
@@ -647,7 +668,7 @@ def display_hotpath(data, fp, limit=0.05, **kwargs):
             if len(stack) > 1:
                 i = 1
                 # Skip boiler plate parts of the stack
-                while i < len(stack) and '%s:%s' % (stack[i].filename(), stack[i].function) in skips:
+                while i < len(stack) and stack[i].skipname() in skips:
                     i += 1
                 if i < len(stack):
                     child.add(stack[i:], time)
@@ -657,6 +678,7 @@ def display_hotpath(data, fp, limit=0.05, **kwargs):
     for sample in data.samples:
         root.add(sample.stack[::-1], sample.time - lasttime)
         lasttime = sample.time
+    showtime = kwargs.get(r'showtime', True)
 
     def _write(node, depth, multiple_siblings):
         site = node.site
@@ -674,10 +696,12 @@ def display_hotpath(data, fp, limit=0.05, **kwargs):
             # lots of string formatting
             listpattern = ''.ljust(indent) +\
                           ('\\' if multiple_siblings else '|') +\
-                          ' %4.1f%%  %s %s'
+                          ' %4.1f%%' +\
+                          (' %5.2fs' % node.count if showtime else '') +\
+                          '  %s %s'
             liststring = listpattern % (node.count / root.count * 100,
                                         filename, function)
-            codepattern = '%' + str(55 - len(liststring)) + 's %s:  %s'
+            codepattern = '%' + ('%d' % (55 - len(liststring))) + 's %d:  %s'
             codestring = codepattern % ('line', site.lineno, site.getsource(30))
 
             finalstring = liststring + codestring
@@ -688,7 +712,7 @@ def display_hotpath(data, fp, limit=0.05, **kwargs):
             # Make frames that didn't actually perform work dark grey
             elif node.count - childrensamples == 0:
                 finalstring = '\033[90m' + finalstring + '\033[0m'
-            print(finalstring, file=fp)
+            fp.write(finalstring + b'\n')
 
         newdepth = depth
         if len(visiblechildren) > 1 or multiple_siblings:
@@ -705,9 +729,8 @@ def write_to_flame(data, fp, scriptpath=None, outputfile=None, **kwargs):
     if scriptpath is None:
         scriptpath = encoding.environ['HOME'] + '/flamegraph.pl'
     if not os.path.exists(scriptpath):
-        print("error: missing %s" % scriptpath, file=fp)
-        print("get it here: https://github.com/brendangregg/FlameGraph",
-              file=fp)
+        fp.write(b'error: missing %s\n' % scriptpath)
+        fp.write(b'get it here: https://github.com/brendangregg/FlameGraph\n')
         return
 
     fd, path = pycompat.mkstemp()
@@ -725,7 +748,7 @@ def write_to_flame(data, fp, scriptpath=None, outputfile=None, **kwargs):
             lines[line] = 1
 
     for line, count in lines.iteritems():
-        file.write("%s %s\n" % (line, count))
+        file.write("%s %d\n" % (line, count))
 
     file.close()
 
@@ -733,7 +756,7 @@ def write_to_flame(data, fp, scriptpath=None, outputfile=None, **kwargs):
         outputfile = '~/flamegraph.svg'
 
     os.system("perl ~/flamegraph.pl %s > %s" % (path, outputfile))
-    print("Written to %s" % outputfile, file=fp)
+    fp.write(b'Written to %s\n' % outputfile)
 
 _pathcache = {}
 def simplifypath(path):
@@ -759,11 +782,18 @@ def write_to_json(data, fp):
         stack = []
 
         for frame in sample.stack:
-            stack.append((frame.path, frame.lineno, frame.function))
+            stack.append(
+                (pycompat.sysstr(frame.path),
+                 frame.lineno,
+                 pycompat.sysstr(frame.function)))
 
         samples.append((sample.time, stack))
 
-    print(json.dumps(samples), file=fp)
+    data = json.dumps(samples)
+    if not isinstance(data, bytes):
+        data = data.encode('utf-8')
+
+    fp.write(data)
 
 def write_to_chrome(data, fp, minthreshold=0.005, maxthreshold=0.999):
     samples = []
@@ -789,9 +819,6 @@ def write_to_chrome(data, fp, minthreshold=0.005, maxthreshold=0.999):
             id2stack[-1].update(parent=parent)
         return myid
 
-    def endswith(a, b):
-        return list(a)[-len(b):] == list(b)
-
     # The sampling profiler can sample multiple times without
     # advancing the clock, potentially causing the Chrome trace viewer
     # to render single-pixel columns that we cannot zoom in on.  We
@@ -831,9 +858,6 @@ def write_to_chrome(data, fp, minthreshold=0.005, maxthreshold=0.999):
     # events given only stack snapshots.
 
     for sample in data.samples:
-        tos = sample.stack[0]
-        name = tos.function
-        path = simplifypath(tos.path)
         stack = tuple((('%s:%d' % (simplifypath(frame.path), frame.lineno),
                         frame.function) for frame in sample.stack))
         qstack = collections.deque(stack)
@@ -861,7 +885,7 @@ def write_to_chrome(data, fp, minthreshold=0.005, maxthreshold=0.999):
     fp.write('\n')
 
 def printusage():
-    print("""
+    print(r"""
 The statprof command line allows you to inspect the last profile's results in
 the following forms:
 
@@ -892,17 +916,17 @@ def main(argv=None):
 
     optstart = 2
     displayargs['function'] = None
-    if argv[1] == 'hotpath':
+    if argv[1] == r'hotpath':
         displayargs['format'] = DisplayFormats.Hotpath
-    elif argv[1] == 'lines':
+    elif argv[1] == r'lines':
         displayargs['format'] = DisplayFormats.ByLine
-    elif argv[1] == 'functions':
+    elif argv[1] == r'functions':
         displayargs['format'] = DisplayFormats.ByMethod
-    elif argv[1] == 'function':
+    elif argv[1] == r'function':
         displayargs['format'] = DisplayFormats.AboutMethod
         displayargs['function'] = argv[2]
         optstart = 3
-    elif argv[1] == 'flame':
+    elif argv[1] == r'flame':
         displayargs['format'] = DisplayFormats.FlameGraph
     else:
         printusage()
@@ -920,22 +944,22 @@ def main(argv=None):
     displayargs['limit'] = 0.05
     path = None
     for o, value in opts:
-        if o in ("-l", "--limit"):
+        if o in (r"-l", r"--limit"):
             displayargs['limit'] = float(value)
-        elif o in ("-f", "--file"):
+        elif o in (r"-f", r"--file"):
             path = value
-        elif o in ("-o", "--output-file"):
+        elif o in (r"-o", r"--output-file"):
             displayargs['outputfile'] = value
-        elif o in ("-p", "--script-path"):
+        elif o in (r"-p", r"--script-path"):
             displayargs['scriptpath'] = value
-        elif o in ("-h", "help"):
+        elif o in (r"-h", r"help"):
             printusage()
             return 0
         else:
             assert False, "unhandled option %s" % o
 
     if not path:
-        print('must specify --file to load')
+        print(r'must specify --file to load')
         return 1
 
     load_data(path=path)
@@ -944,5 +968,5 @@ def main(argv=None):
 
     return 0
 
-if __name__ == "__main__":
+if __name__ == r"__main__":
     sys.exit(main())
